import logging
import os
import time
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from numpyencoder import NumpyEncoder
from pycaret.classification import *
from sklearn import metrics
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from xgboost import XGBClassifier

from MEDiml.learning.DataCleaner import DataCleaner
from MEDiml.learning.DesignExperiment import DesignExperiment
from MEDiml.learning.FSR import FSR
from MEDiml.learning.ml_utils import (average_results, combine_rad_tables,
                                        feature_imporance_analysis,
                                        finalize_rad_table, get_ml_test_table,
                                        get_radiomics_table, intersect,
                                        intersect_var_tables, save_model)
from MEDiml.learning.Normalization import Normalization
from MEDiml.learning.Results import Results

from ..utils.json_utils import load_json, save_json


class RadiomicsLearner:
    def __init__(self, path_study: Path, path_settings: Path, experiment_label: str) -> None:
        """
        Constructor of the class DesignExperiment.

        Args:
            path_study (Path): Path to the main study folder where the outcomes, 
                learning patients and holdout patients dictionaries are found.
            path_settings (Path): Path to the settings folder.
            experiment_label (str): String specifying the label to attach to a given learning experiment in 
                "path_experiments". This label will be attached to the ml__$experiments_label$.json file as well
                as the learn__$experiment_label$ folder. This label is used to keep track of different experiments 
                with different settings (e.g. radiomics, scans, machine learning algorithms, etc.).
        
        Returns:
            None
        """
        self.path_study = Path(path_study)
        self.path_settings = Path(path_settings)
        self.experiment_label = experiment_label
    
    def __load_ml_info(self, ml_dict_paths: Dict) -> Dict:
        """
        Initializes the test dictionary information (training patients, test patients, ML dict, etc).

        Args:
            ml_dict_paths (Dict): Dictionary containing the paths to the different files needed 
                to run the machine learning experiment.
        
        Returns:
            dict: Dictionary containing the information of the machine learning test.
        """
        ml_dict = dict()

        # Training and test patients
        ml_dict['patientsTrain'] = load_json(ml_dict_paths['patientsTrain'])
        ml_dict['patientsTest'] = load_json(ml_dict_paths['patientsTest'])

        # Outcome table for training and test patients
        outcome_table = pd.read_csv(ml_dict_paths['outcomes'], index_col=0)
        ml_dict['outcome_table_binary'] = outcome_table.iloc[:, [0]]
        if outcome_table.shape[1] == 2:
            ml_dict['outcome_table_time'] = outcome_table.iloc[:, [1]]
        
        # Machine learning dictionary
        ml_dict['ml'] = load_json(ml_dict_paths['ml'])
        ml_dict['path_results'] = ml_dict_paths['results']

        return ml_dict

    def __find_balanced_threshold(
            self, 
            model: XGBClassifier, 
            variable_table: pd.DataFrame, 
            outcome_table_binary: pd.DataFrame
        ) -> float:
        """
        Finds the balanced threshold for the given machine learning test.

        Args:
            model (XGBClassifier): Trained XGBoost classifier for the given machine learning run.
            variable_table (pd.DataFrame): Radiomics table.
            outcome_table_binary (pd.DataFrame): Outcome table with binary labels.
        
        Returns:
            float: Balanced threshold for the given machine learning test.
        """
        # Check is there is a feature mismatch
        if model.feature_names_in_.shape[0] != variable_table.columns.shape[0]:
            variable_table = variable_table.loc[:, model.feature_names_in_]

        # Getting the probability responses for each patient
        prob_xgb = np.zeros((variable_table.index.shape[0], 1)) * np.nan
        patient_ids = list(variable_table.index.values)
        for p in range(variable_table.index.shape[0]):
            prob_xgb[p] = self.predict_xgb(model, variable_table.loc[[patient_ids[p]], :])

        # Calculating the ROC curve
        fpr, tpr, thresholds = metrics.roc_curve(outcome_table_binary.iloc[:, 0], prob_xgb)

        # Calculating the optimal threshold by minizing fpr (false positive rate) and maximizing tpr (true positive rate)
        minimum = np.argmin(np.power(fpr, 2) + np.power(1-tpr, 2))
        
        return thresholds[minimum]
    
    def get_hold_out_set_table(self, ml: Dict, var_id: str, patients_id: List):
        """
        Loads and pre-processes different radiomics tables then combines them to be used for hold-out testing.

        Args:
            ml (Dict): The machine learning dictionary containing the information of the machine learning test.
            var_id (str): String specifying the ID of the radiomics variable in ml.
                --> Ex: var1
            patients_id (List): List of patients of the hold-out set.

        Returns:
            pd.DataFrame: Radiomics table for the hold-out set.
        """
        # Loading and pre-processing
        rad_var_struct = ml['variables'][var_id]
        rad_tables_holdout = list()
        for item in rad_var_struct['path'].values():
            # Reading the table
            path_radiomics_csv = item['csv']
            path_radiomics_txt = item['txt']
            image_type = item['type']
            rad_table_holdout = get_radiomics_table(path_radiomics_csv, path_radiomics_txt, image_type, patients_id)
            rad_tables_holdout.append(rad_table_holdout)
        
        # Combine the tables
        rad_tables_holdout = combine_rad_tables(rad_tables_holdout)
        rad_tables_holdout.Properties['userData']['flags_processing'] = {}

        return rad_tables_holdout
    
    def pre_process_variables(self, ml: Dict, outcome_table_binary: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Loads and pre-processes different radiomics tables from different variable types
        found in the ml dict.
        
        Note: 
            only patients of the training/learning set should be found in this outcome table.

        Args:
            ml (Dict): The machine learning dictionary containing the information of the machine learning test.
            outcome_table_binary (pd.DataFrame): outcome table with binary labels. This table may be used to
                pre-process some variables with the "FDA" feature set reduction algorithm.

        Returns:
            Tuple: Two dict of processed radiomics tables, one dict for training and one for 
                testing (no feature set reduction). 
        """
        # Get a list of unique variables found in the ml variables combinations dict
        variables_id = [s.split('_') for s in ml['variables']['combinations']]
        variables_id = list(set([x for sublist in variables_id for x in sublist]))

        # For each variable, load the corresponding radiomics table and pre-process it
        processed_var_tables, processed_var_tables_test =  {var_id : self.pre_process_radiomics_table(
            ml, 
            var_id, 
            outcome_table_binary
        ) for var_id in variables_id}
        
        return processed_var_tables, processed_var_tables_test

    def pre_process_radiomics_table(
            self, 
            ml: Dict, 
            var_id: str, 
            outcome_table_binary: pd.DataFrame,
            patients_train: list
        ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        For the given variable, this function loads the corresponding radiomics tables and pre-processes them
        (cleaning, normalization and feature set reduction).

        Note: 
            Only patients of the training/learning set should be found in the given outcome table.
        
        Args:
            ml (Dict): The machine learning dictionary containing the information of the machine learning test 
                (parameters, options, etc.).
            var_id (str): String specifying the ID of the radiomics variable in ml. For example: 'var1'.
            outcome_table_binary (pd.DataFrame): outcome table with binary labels. This table may
                be used to pre-process some variables with the "FDA" feature set reduction algorithm.
            
            patients_train (list): List of patients to use for training.
        
        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]: Two dataframes of processed radiomics tables, one for training 
                and one for testing (no feature set reduction).
        """
        # Initialization
        patient_ids = list(outcome_table_binary.index)
        outcome_table_binary_training = outcome_table_binary.loc[patients_train]
        var_names = ['var_datacleaning', 'var_normalization', 'var_fSetReduction']
        flags_preprocessing =  {key: key in ml['variables'][var_id].keys() for key in var_names}
        flags_preprocessing_test = flags_preprocessing.copy()
        flags_preprocessing_test['var_fSetReduction'] = False

        # Pre-processing
        rad_var_struct = ml['variables'][var_id]
        rad_tables_learning = list()
        for item in rad_var_struct['path'].values():
            # Loading the table
            path_radiomics_csv = item['csv']
            path_radiomics_txt = item['txt']
            image_type = item['type']
            rad_table_learning = get_radiomics_table(path_radiomics_csv, path_radiomics_txt, image_type, patient_ids)

            # Data cleaning
            if flags_preprocessing['var_datacleaning']:
                cleaning_dict = ml['datacleaning'][ml['variables'][var_id]['var_datacleaning']]['feature']['continuous']
                data_cleaner = DataCleaner(rad_table_learning)
                rad_table_learning = data_cleaner(cleaning_dict)
                if rad_table_learning is None:
                    continue

            # Normalization (ComBat)
            if flags_preprocessing['var_normalization']:
                normalization_method = ml['variables'][var_id]['var_normalization']
                # Some information must be stored to re-apply combat for testing data
                if 'combat' in normalization_method.lower():
                    # Training data
                    rad_table_learning.Properties['userData']['normalization'] = dict()
                    rad_table_learning.Properties['userData']['normalization']['original_data'] = dict()
                    rad_table_learning.Properties['userData']['normalization']['original_data']['path_radiomics_csv'] = path_radiomics_csv
                    rad_table_learning.Properties['userData']['normalization']['original_data']['path_radiomics_txt'] = path_radiomics_txt
                    rad_table_learning.Properties['userData']['normalization']['original_data']['image_type'] = image_type
                    rad_table_learning.Properties['userData']['normalization']['original_data']['patient_ids'] = patient_ids
                    if flags_preprocessing['var_datacleaning']:
                        data_cln_method = ml['variables'][var_id]['var_datacleaning']
                        rad_table_learning.Properties['userData']['normalization']['original_data']['datacleaning_method'] = data_cln_method
                    
                    # Apply ComBat
                    normalization = Normalization('combat')
                    rad_table_learning = normalization.apply_combat(variable_table=rad_table_learning)  # Training data
                else:
                    raise NotImplementedError(f'Normalization method: {normalization_method} not recognized.')

            # Save the table
            rad_tables_learning.append(rad_table_learning)

        # Seperate training and testing data before feature set reduction
        rad_tables_testing = deepcopy(rad_tables_learning)
        rad_tables_training = []
        for rad_tab in rad_tables_learning:
            patients_ids = intersect(patients_train, list(rad_tab.index))
            rad_tables_training.append(deepcopy(rad_tab.loc[patients_ids]))

        # Deepcopy properties
        temp_properties = list()
        for rad_tab in rad_tables_testing:
            temp_properties.append(deepcopy(rad_tab.Properties))

        # Feature set reduction (for training data only)
        if flags_preprocessing['var_fSetReduction']:
            f_set_reduction_method = ml['variables'][var_id]['var_fSetReduction']['method']
            fsr = FSR(f_set_reduction_method)
            
            # Apply FDA
            rad_tables_training = fsr.apply_fsr(
                ml, 
                rad_tables_training, 
                outcome_table_binary_training, 
                path_save_logging=ml['path_results']
            )

        # Re-assign properties
        for i in range(len(rad_tables_testing)):
            rad_tables_testing[i].Properties = temp_properties[i]
        del temp_properties
        
        # Finalization steps
        rad_tables_training.Properties['userData']['flags_preprocessing'] = flags_preprocessing
        rad_tables_testing = combine_rad_tables(rad_tables_testing)
        rad_tables_testing.Properties['userData']['flags_processing'] = flags_preprocessing_test

        return rad_tables_training, rad_tables_testing

    def train_xgboost_model(
            self, 
            var_table_train: pd.DataFrame,
            outcome_table_binary_train: pd.DataFrame,
            var_importance_threshold: float = 0.05,
            optimal_threshold: float = None,
            optimization_metric: str = 'MCC',
            method : str = "pycaret",
            use_gpu: bool = False,
            seed: int = None,
        ) -> Dict:
        """
        Trains an XGBoost model for the given machine learning test.

        Args:
            var_table_train (pd.DataFrame): Radiomics table for the training/learning set.
            outcome_table_binary_train (pd.DataFrame): Outcome table with binary labels for the training/learning set.
            var_importance_threshold (float): Threshold for the variable importance. Variables with importance below
                this threshold will be removed from the model.
            optimal_threshold (float, optional): Optimal threshold for the XGBoost model. If not given, it will be
                computed using the training set.
            optimization_metric (str, optional): String specifying the metric to use to optimize the ml model.
            method (str, optional): String specifying the method to use to train the XGBoost model.
                - "pycaret": Use PyCaret to train the model (automatic).
                - "grid_search": Grid search with cross-validation to find the best parameters.
                - "random_search": Random search with cross-validation to find the best parameters.
            use_gpu (bool, optional): Boolean specifying if the GPU should be used to train the model. Default is True.
            seed (int, optional): Integer specifying the seed to use for the random number generator.
        
        Returns:
            Dict: Dictionary containing info about the trained XGBoost model.
        """
        
        # Safety check (make sure that the outcome table and the variable table have the same patients)
        var_table_train, outcome_table_binary_train = intersect_var_tables(var_table_train, outcome_table_binary_train)

        # Finalize the new radiomics table with the remaining variables
        var_table_train = finalize_rad_table(var_table_train)

        if method.lower() == "pycaret":
            # Set up data for PyCaret
            temp_data = pd.merge(var_table_train, outcome_table_binary_train, left_index=True, right_index=True)

            # PyCaret setup
            setup(
                data=temp_data,
                feature_selection=True,
                n_features_to_select=1-var_importance_threshold,
                fold=5,
                target=temp_data.columns[-1],
                use_gpu=use_gpu,
                feature_selection_estimator="xgboost",
                session_id=seed
            )

            # Set seed
            if seed is not None:
                set_config('seed', seed)

            # Creating XGBoost model using PyCaret
            classifier = create_model('xgboost', verbose=False)

            # Tuning XGBoost model using PyCaret
            classifier = tune_model(classifier, optimize=optimization_metric)
        
        else:
            # Initial training to filter features using variable importance
            # XGB Classifier
            classifier = XGBClassifier()
            classifier.fit(var_table_train, outcome_table_binary_train)
            var_importance = classifier.feature_importances_

            # Normalize var_importance if necessary
            if np.sum(var_importance) != 1:
                var_importance_threshold = var_importance_threshold / np.sum(var_importance)
                var_importance = var_importance / np.sum(var_importance)

            # Filter variables
            var_table_train = var_table_train.iloc[:, var_importance >= var_importance_threshold]

            # Check if variable table is empty after filtering
            if var_table_train.shape[1] == 0:
                raise ValueError('Variable table is empty after variable importance filtering. Use a smaller threshold.')

            # Suggested scale_pos_weight
            scale_pos_weight = 1 - (outcome_table_binary_train == 0).sum().values[0] \
                / (outcome_table_binary_train == 1).sum().values[0]

            # XGB Classifier
            classifier = XGBClassifier(scale_pos_weight=scale_pos_weight)

            # Tune XGBoost parameters
            params = {
                'max_depth': [3, 4, 5], 
                'learning_rate': [0.1 , 0.01, 0.001], 
                'n_estimators': [50, 100, 200]
            }

            if method.lower() == "grid_search":
                # Set up grid search with cross-validation
                grid_search = GridSearchCV(
                    estimator=classifier, 
                    param_grid=params, 
                    cv=5, 
                    n_jobs=-1, 
                    verbose=3, 
                    scoring='matthews_corrcoef'
                )
            elif method.lower() == "random_search":
                # Set up random search with cross-validation
                grid_search = RandomizedSearchCV(
                    estimator=classifier, 
                    param_distributions=params, 
                    cv=5, 
                    n_jobs=-1, 
                    verbose=3, 
                    scoring='matthews_corrcoef'
                )
            else:
                raise NotImplementedError(f'Method: {method} not recognized. Use "grid_search", "random_search", "auto" or "pycaret".')
            
            # Fit the grid search
            grid_search.fit(var_table_train, outcome_table_binary_train)

            # Get the best parameters
            best_params = grid_search.best_params_

            # Fit the XGB Classifier with the best parameters
            classifier = XGBClassifier(**best_params)
            classifier.fit(var_table_train, outcome_table_binary_train)
        
        # Saving the information of the model in a dictionary
        model_xgb = dict()
        model_xgb['algo'] = 'xgb'
        model_xgb['type'] = 'binary'
        model_xgb['method'] = method
        if optimal_threshold:
            model_xgb['threshold'] = optimal_threshold
        else:
            try:
                model_xgb['threshold'] = self.__find_balanced_threshold(classifier, var_table_train, outcome_table_binary_train)
            except Exception as e:
                print('Error in finding optimal threshold, it will be set to 0.5:' + str(e))
                model_xgb['threshold'] = 0.5
        model_xgb['model'] = classifier
        model_xgb['var_names'] = list(classifier.feature_names_in_)
        model_xgb['var_info'] = deepcopy(var_table_train.Properties['userData'])
        if method == "auto":
            model_xgb['optimization'] = "auto"
        elif method == "pycaret":
            model_xgb['optimization'] = classifier.get_params()
        else:
            model_xgb['optimization'] = best_params

        return model_xgb
        
    def test_xgb_model(self, model_dict: Dict, variable_table: pd.DataFrame, patient_list: List) -> List:
        """
        Tests the XGBoost model for the given dataset patients.

        Args:
            model_dict (Dict): Dictionary containing info about the trained XGBoost model.
            variable_table (pd.DataFrame): Radiomics table for the test set (should not be normalized).
            patient_list (List): List of patients to test.
        
        Returns:
            List: List the model response for the training and test sets.
        """
        # Initialization
        n_test = len(patient_list)
        var_names = model_dict['var_names']
        var_def = model_dict['var_info']['variables']['var_def']
        model_response = list()

        # Preparing the variable table
        variable_table = get_ml_test_table(variable_table, var_names, var_def)

        # Test the model
        for i in range(n_test):
            # Get the patient IDs
            patient_ids = patient_list[i]

            # Getting predictions for each patient
            n_patients = len(patient_ids)
            varargout = np.zeros((n_patients, 1)) * np.nan  # NaN if the computation fails            
            for p in range(n_patients):
                try:
                    varargout[p] = self.predict_xgb(model_dict['model'], variable_table.loc[[patient_ids[p]], :])
                except Exception as e:
                    print('Error in computing prediction for patient ' + str(patient_ids[p]) + ': ' + str(e))
                    varargout[p] = np.nan

            # Save the predictions 
            model_response.append(varargout)

        return model_response
    
    def predict_xgb(self, xgb_model: XGBClassifier, variable_table: pd.DataFrame) -> float:
        """
        Computes the prediction of the XGBoost model for the given variable table.

        Args:
            xgb_model (XGBClassifier): XGBClassifier model.
            variable_table (pd.DataFrame): Variable table for the prediction.
        
        Returns:
            float: Prediction of the XGBoost model.
        """

        # Predictions
        predictions = xgb_model.predict_proba(variable_table)

        # Get the probability of the positive class
        predictions = predictions[:, 1][0]

        return predictions

    def ml_run(self, path_ml: Path, holdout_test: bool = True, method: str = 'auto') -> None:
        """
        This function runs the machine learning test for the ceated experiment.

        Args:
            path_ml (Path): Path to the main dictionary containing info about the ml current experiment.
            holdout_test (bool, optional): Boolean specifying if the hold-out test should be performed.
        
        Returns:
            None.
        """
        # Set up logging file for the batch
        log_file = os.path.dirname(path_ml) + '/batch.log'
        logging.basicConfig(filename=log_file, level=logging.INFO, format='%(message)s', filemode='w')

        # Start the timer
        batch_start = time.time()

        logging.info("\n\n********************MACHINE LEARNING RUN********************\n\n")

        # --> A. Initialization phase
        # Load the test dictionary and machine learning information
        ml_dict_paths = load_json(path_ml)      # Test information dictionary
        ml_info_dict = self.__load_ml_info(ml_dict_paths)       # Machine learning information dictionary

        # Machine learning assets
        patients_train = ml_info_dict['patientsTrain']
        patients_test = ml_info_dict['patientsTest']
        patients_holdout = load_json(self.path_study / 'patientsHoldOut.json') if holdout_test else None
        outcome_table_binary = ml_info_dict['outcome_table_binary']
        ml = ml_info_dict['ml']
        path_results = ml_info_dict['path_results']
        ml['path_results'] = path_results

        # --> B. Machine Learning phase 
        # B.1. Pre-processing features
        start = time.time()
        logging.info("\n\n--> PRE-PROCESSING TRAINING VARIABLES")

        # Not all variables will be used to train the model, only the user-selected variable
        var_id = str(ml['variables']['varStudy'])

        # Pre-processing of the radiomics tables/variables
        processed_training_table, processed_testing_table = self.pre_process_radiomics_table(
            ml, 
            var_id, 
            outcome_table_binary.copy(),
            patients_train
        )
        logging.info(f"...Done in {time.time()-start} s")

        # B.2. Pre-learning initialization
        # Patient definitions (training and test sets)
        patient_ids = list(outcome_table_binary.index)
        patients_train = intersect(intersect(patient_ids, patients_train), processed_training_table.index)
        patients_test = intersect(intersect(patient_ids, patients_test), processed_testing_table.index)
        patients_holdout = intersect(patient_ids, patients_holdout) if holdout_test else None

        # Initializing outcome tables for training and test sets
        outcome_table_binary_train = outcome_table_binary.loc[patients_train, :]
        outcome_table_binary_test = outcome_table_binary.loc[patients_test, :]
        outcome_table_binary_holdout = outcome_table_binary.loc[patients_holdout, :] if holdout_test else None

        # Serperate variable table for training sets (repetitive but double-checking)
        var_table_train = processed_training_table.loc[patients_train, :]

        # Initializing XGBoost model settings
        var_importance_threshold = ml['algorithms']['XGBoost']['varImportanceThreshold']
        optimal_threshold = ml['algorithms']['XGBoost']['optimalThreshold']
        optimization_metric = ml['algorithms']['XGBoost']['optimizationMetric']
        method = ml['algorithms']['XGBoost']['method'] if 'method' in ml['algorithms']['XGBoost'].keys() else method
        use_gpu = ml['algorithms']['XGBoost']['useGPU'] if 'useGPU' in ml['algorithms']['XGBoost'].keys() else True
        seed = ml['algorithms']['XGBoost']['seed'] if 'seed' in ml['algorithms']['XGBoost'].keys() else None

        # B.2. Training the XGBoost model
        tstart = time.time()
        logging.info(f"\n\n--> TRAINING XGBOOST MODEL FOR VARIABLE {var_id}")

        # Training the model
        model = self.train_xgboost_model(
            var_table_train, 
            outcome_table_binary_train, 
            var_importance_threshold, 
            optimal_threshold,
            method=method,
            use_gpu=use_gpu,
            optimization_metric=optimization_metric,
            seed=seed
        )

        # Saving the trained model using pickle
        name_save_model = ml['algorithms']['XGBoost']['nameSave']
        model_id = name_save_model + '_' + str(ml['variables']['varStudy'])
        path_model = os.path.dirname(path_results) + '/' + (model_id + '.pickle')
        model_dict = save_model(model, str(ml['variables']['varStudy']), path_model, ml=ml)

        logging.info("{}--> DONE. TOTAL TIME OF LEARNING PROCESS: {:.2f} min".format(" " * 4, (time.time()-tstart) / 60))

        # --> C. Testing phase        
        # C.1. Testing the XGBoost model and computing model response
        tstart = time.time()
        logging.info(f"\n\n--> TESTING XGBOOST MODEL FOR VARIABLE {var_id}")

        response_train, response_test = self.test_xgb_model(
            model,
            processed_testing_table,
            [patients_train, patients_test]
        )
        
        logging.info('{}--> DONE. TOTAL TIME OF LEARNING PROCESS: {:.2f}'.format(" " * 4, (time.time() - tstart)/60))
        
        if holdout_test:
            # --> D. Holdoutset testing phase
            # D.1. Prepare holdout test data
            var_table_all_holdout = self.get_hold_out_set_table(ml, var_id, patients_holdout)

            # D.2. Testing the XGBoost model and computing model response on the holdout set
            tstart = time.time()
            logging.info(f"\n\n--> TESTING XGBOOST MODEL FOR VARIABLE {var_id} ON THE HOLDOUT SET")

            response_holdout = self.test_xgb_model(model, var_table_all_holdout, [patients_holdout])[0]
        
        logging.info('{}--> DONE. TOTAL TIME OF LEARNING PROCESS: {:.2f}'.format(" " * 4, (time.time() - tstart)/60))
        
        # E. Computing performance metrics
        tstart = time.time()

        # Initialize the Results class
        result = Results(model_dict, model_id)
        if holdout_test:
            run_results = result.to_json(
                response_train=response_train, 
                response_test=response_test,
                response_holdout=response_holdout, 
                patients_train=patients_train, 
                patients_test=patients_test, 
                patients_holdout=patients_holdout
            )
        else:
            run_results = result.to_json(
                response_train=response_train, 
                response_test=response_test,
                response_holdout=None, 
                patients_train=patients_train, 
                patients_test=patients_test, 
                patients_holdout=None
            )
        
        # Calculating performance metrics for training phase and saving the ROC curve
        run_results[model_id]['train']['metrics'] = result.get_model_performance(
            response_train, 
            outcome_table_binary_train,
        )
        
        # Calculating performance metrics for testing phase and saving the ROC curve
        run_results[model_id]['test']['metrics'] = result.get_model_performance(
            response_test, 
            outcome_table_binary_test,
        )

        if holdout_test:
            # Calculating performance metrics for holdout phase and saving the ROC curve
            run_results[model_id]['holdout']['metrics'] = result.get_model_performance(
                response_holdout, 
                outcome_table_binary_holdout,
            )

        logging.info('\n\n--> COMPUTING PERFORMANCE METRICS ... Done in {:.2f} sec'.format(time.time()-tstart))
        
        # F. Saving the results dictionary
        save_json(path_results, run_results, cls=NumpyEncoder)

        # Total computing time
        logging.info("\n\n*********************************************************************")
        logging.info('{} TOTAL COMPUTATION TIME: {:.2f} hours'.format(" " * 13, (time.time()-batch_start)/3600))
        logging.info("*********************************************************************")
        
    def run_experiment(self, holdout_test: bool = True, method: str = "pycaret") -> None:
        """
        Run the machine learning experiment for each split/run

        Args:
            holdout_test (bool, optional): Boolean specifying if the hold-out test should be performed.
            method (str, optional): String specifying the method to use to train the XGBoost model.
                - "pycaret": Use PyCaret to train the model (automatic).
                - "grid_search": Grid search with cross-validation to find the best parameters.
                - "random_search": Random search with cross-validation to find the best parameters.
            
        Returns:
            None
        """
        # Initialize the DesignExperiment class
        experiment = DesignExperiment(self.path_study, self.path_settings, self.experiment_label)

        # Generate the machine learning experiment
        path_file_ml_paths = experiment.generate_experiment()

        # Run the different machine learning tests for the experiment
        tests_dict = load_json(path_file_ml_paths) # Tests dictionary
        for run in tests_dict.keys():
            self.ml_run(tests_dict[run], holdout_test, method)
        
        # Average results of the different splits/runs
        average_results(self.path_study / f'learn__{self.experiment_label}', save=True)

        # Analyze the features importance for all the runs
        feature_imporance_analysis(self.path_study / f'learn__{self.experiment_label}')
        


import math
from itertools import combinations, permutations
from typing import List, Union

import numpy as np
import pywt

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj


class Wavelet():
    """
    The wavelet filter class.
    """

    def __init__(
            self,
            ndims: int,
            wavelet_name="haar",
            padding="symmetric",
            rot_invariance=False):
        """The constructor of the wavelet filter

        Args:
            ndims (int): The number of dimension of the images that will be filter as int.
            wavelet_name (str): The name of the wavelet kernel as string.
            padding (str): The padding type that will be used to produce the convolution
            rot_invariance (bool): If true, rotation invariance will be done on the images.

        Returns:
            None
        """
        self.dim = ndims
        self.padding = padding
        self.rot = rot_invariance
        self.wavelet = None
        self.kernel_length = None
        self.create_kernel(wavelet_name)

    def create_kernel(self,
                      wavelet_name: str):
        """Get the wavelet object and his kernel length.

        Args:
            wavelet_name (str): A string that represent the wavelet name that will be use to create the kernel

        Returns:
            None
        """

        self.wavelet = pywt.Wavelet(wavelet_name)
        self.kernel_length = max(self.wavelet.rec_len, self.wavelet.dec_len)

    def __unpad(self,
                images: np.ndarray,
                padding: List) -> np.ndarray:
        """Unpad a batch of images

        Args:
            images: A numpy nd-array or a list that represent the batch of padded images.
                    The shape should be (B, H, W) or (B, H, W, D)
            padding: a list of length 2*self.dim that gives the length of padding on each side of each axis.

        Returns: 
            ndarray: A numpy nd-array or a list that represent the batch of unpadded images
        """

        if self.dim == 2:
            return images[:, padding[0]:-padding[1], padding[2]:-padding[3]]
        elif self.dim == 3:
            return images[:, padding[0]:-padding[1], padding[2]:-padding[3], padding[4]:-padding[5]]
        else:
            raise NotImplementedError

    def __get_pad_length(self,
                         image_shape: List,
                         level: int) -> np.ndarray:
        """Compute the padding length needed to have a padded image where the length
        along each axis is a multiple 2^level.

        Args:
            image_shape (List): a list of integer that describe the length of the image along each axis.
            level (int): The level of the wavelet transform

        Returns: 
            ndarray: An integer list of length 2*self.dim that gives the length of padding on each side of each axis.
        """
        padding = []
        ker_length = self.kernel_length*level
        for l in image_shape:
            padded_length = math.ceil((l + 2*(ker_length-1)) / 2**level) * 2**level - l
            padding.extend([math.floor(padded_length/2), math.ceil(padded_length/2)])

        return padding

    def _pad_imgs(self,
                  images: np.ndarray,
                  padding,
                  axis: List):
        """Apply padding on a 3d images using a 2D padding pattern (special for wavelet).

        Args:
            images: a numpy array that represent the image.
            padding: The padding length that will apply on each side of each axe.
            axis: A list of axes on which the padding will be done.

        Returns: 
            ndarray: A numpy array that represent the padded image.
        """
        pad_tuple = ()
        j = 0

        for i in range(np.ndim(images)):
            if i in axis:
                pad_tuple += ((padding[j], padding[j+1]),)
                j += 2
            else:
                pad_tuple += ((0, 0),)

        return np.pad(images, pad_tuple, mode=self.padding)
        

    def convolve(self,
                 images: np.ndarray,
                 _filter="LHL",
                 level=1)-> np.ndarray:
        """Filter a given batch of images using pywavelet.

        Args:
            images (ndarray): A n-dimensional numpy array that represent the images to filter
            _filter (str): The filter to uses.
            level (int): The number of decomposition steps to perform.

        Returns:
            ndarray: The filtered image as numpy nd-array
        """

        # We pad the images
        padding = self.__get_pad_length(np.shape(images[0]), level)
        axis_list = [i for i in range(0, self.dim)]
        images = np.expand_dims(self._pad_imgs(images[0], padding, axis_list), axis=0)

        # We generate the to collect the result from pywavelet dictionary
        _index = str().join(['a' if _filter[i] == 'L' else 'd' for i in range(len(_filter))])

        if self.rot:
            result = []
            _index_list = np.unique([str().join(perm) for perm in permutations(_index, self.dim)])

            # For each images, we flip each axis.
            for image in images:
                axis_rot = [comb for j in range(self.dim+1) for comb in combinations(np.arange(self.dim), j)]
                images_rot = [np.flip(image, axis) for axis in axis_rot]

                res_rot = []
                for i in range(len(images_rot)):
                    filtered_image = pywt.swtn(images_rot[i], self.wavelet, level=level)[0]
                    res_rot.extend([np.flip(filtered_image[j], axis=axis_rot[i]) for j in _index_list])

                result.extend([np.mean(res_rot, axis=0)])
        else:
            result = []
            for i in range(len(images)):
                result.extend([pywt.swtn(images[i], self.wavelet, level=level)[level-1][_index]])

        return self.__unpad(np.array(result), padding)

def apply_wavelet(
        input_images: Union[np.ndarray, image_volume_obj],
        medscan: MEDscan = None,
        ndims: int = 3,
        wavelet_name: str = "haar",
        subband: str = "LHL",
        level: int = 1,
        padding: str = "symmetric",
        rot_invariance: bool = False
    ) -> np.ndarray:
    """Apply the mean filter to the input image
    
    Args:
        input_images (ndarray): The image to filter.
        medscan (MEDscan, optional): The MEDscan object that will provide the filter parameters.
        ndims (int, optional): The number of dimensions of the input image.
        wavelet_name (str): The name of the wavelet kernel as string.
        level (List[str], optional): The number of decompositions steps to perform.
        subband (str, optional): String of the 1D wavelet kernels ("H" for high-pass 
            filter or "L" for low-pass filter). Must have a size of ``ndims``.
        padding (str, optional): The padding type that will be used to produce the convolution. Check options 
            here: `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.
        rot_invariance (bool, optional): If true, rotation invariance will be done on the kernel.

    Returns:
        ndarray: The filtered image.
    """
    # Check if the input is a numpy array or a Image volume object
    spatial_ref = None
    if type(input_images) == image_volume_obj:
        spatial_ref = input_images.spatialRef
        input_images = input_images.data
    
    # Convert to shape : (B, W, H, D)
    input_images = np.expand_dims(input_images.astype(np.float64), axis=0) 

    if medscan:
        # Initialize filter class instance
        _filter = Wavelet(
                        ndims=medscan.params.filter.wavelet.ndims, 
                        wavelet_name=medscan.params.filter.wavelet.basis_function,
                        rot_invariance=medscan.params.filter.wavelet.rot_invariance,
                        padding=medscan.params.filter.wavelet.padding
                    )
        # Run convolution
        result = _filter.convolve(
                        input_images, 
                        _filter=medscan.params.filter.wavelet.subband, 
                        level=medscan.params.filter.wavelet.level
                    )
    else:
        # Initialize filter class instance
        _filter = Wavelet(
                        ndims=ndims, 
                        wavelet_name=wavelet_name,
                        rot_invariance=rot_invariance,
                        padding=padding
                    )
        # Run convolution
        result = _filter.convolve(
                        input_images, 
                        _filter=subband, 
                        level=level
                    )
    
    if spatial_ref:
        return image_volume_obj(np.squeeze(result), spatial_ref)
    else:
        return np.squeeze(result)


import pickle
from pathlib import Path

from ..MEDscan import MEDscan


def save_MEDscan(medscan: MEDscan,
                  path_save: Path) -> str:
    """Saves MEDscan class instance in a pickle object
    
    Args:
        medscan (MEDscan): MEDscan instance
        path_save (Path): MEDscan instance saving paths
    
    Returns:
        None.
    """

    series_description = medscan.series_description.translate({ord(ch): '-' for ch in '/\\ ()&:*'})
    name_id = medscan.patientID
    name_id = name_id.translate({ord(ch): '-' for ch in '/\\ ()&:*'})

    # final saving name
    name_complete = name_id + '__' + series_description + '.' + medscan.type + '.npy'
    
    # save
    with open(path_save / name_complete,'wb') as f: 
        pickle.dump(medscan, f)

    return name_complete



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def get_patient_id_from_scan_name(rad_name: str) -> str:
    """
    Finds the patient id from the given string

    Args:
        rad_name(str): Name of a scan or a radiomics structure
    
    Returns:
        str: patient id

    Example:
        >>> get_patient_id_from_scan_name('STS-McGill-001__T1(tumourAndEdema).MRscan')
        STS-McGill-001
    """
    ind_double_under = rad_name.find('__')
    patientID = rad_name[:ind_double_under]

    return patientID



import csv
import json
import os
import pickle
import re
import string
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas
import pandas as pd
import seaborn as sns
from numpyencoder import NumpyEncoder
from sklearn.model_selection import StratifiedKFold

from MEDiml.utils import get_institutions_from_ids
from MEDiml.utils.get_full_rad_names import get_full_rad_names
from MEDiml.utils.json_utils import load_json, save_json


# Define useful constants
# Metrics to process
list_metrics = [
    'AUC', 'AUPRC', 'BAC', 'Sensitivity', 'Specificity',
    'Precision', 'NPV', 'F1_score', 'Accuracy', 'MCC',
    'TN', 'FP', 'FN', 'TP'
]

def average_results(path_results: Path, save: bool = False) -> None:
    """
    Averages the results (AUC, BAC, Sensitivity and Specifity) of all the runs of the same experiment,
    for training, testing and holdout sets.

    Args:
        path_results(Path): path to the folder containing the results of the experiment.
        save (bool, optional): If True, saves the results in the same folder as the model.
    
    Returns:
        None.
    """
    # Get all tests paths
    list_path_tests =  [path for path in path_results.iterdir() if path.is_dir()]

    # Initialize dictionaries
    results_avg = {
        'train': {},
        'test': {},
        'holdout': {}
    }

    # Metrics to process
    metrics = ['AUC', 'AUPRC', 'BAC', 'Sensitivity', 'Specificity',
            'Precision', 'NPV', 'F1_score', 'Accuracy', 'MCC',
            'TN', 'FP', 'FN', 'TP']

    # Process metrics
    for dataset in ['train', 'test', 'holdout']:
        dataset_dict = results_avg[dataset]
        for metric in metrics:
            metric_values = []
            for path_test in list_path_tests:
                results_dict = load_json(path_test / 'run_results.json')
                if dataset in results_dict[list(results_dict.keys())[0]].keys():
                    if 'metrics' in results_dict[list(results_dict.keys())[0]][dataset].keys():
                        metric_values.append(results_dict[list(results_dict.keys())[0]][dataset]['metrics'][metric])
                    else:
                        continue
                else:
                    continue

            # Fill the dictionary
            if metric_values:
                dataset_dict[f'{metric}_mean'] = np.nanmean(metric_values)
                dataset_dict[f'{metric}_std'] = np.nanstd(metric_values)
                dataset_dict[f'{metric}_max'] = np.nanmax(metric_values)
                dataset_dict[f'{metric}_min'] = np.nanmin(metric_values)
                dataset_dict[f'{metric}_2.5%'] = np.nanpercentile(metric_values, 2.5)
                dataset_dict[f'{metric}_97.5%'] = np.nanpercentile(metric_values, 97.5)

    # Save the results
    if save:
        save_json(path_results / 'results_avg.json', results_avg, cls=NumpyEncoder)
        return path_results / 'results_avg.json'

    return results_avg

def combine_rad_tables(rad_tables: List) -> pd.DataFrame:
    """
    Combines a list of radiomics tables into one single table.
    
    Args:
        rad_tables (List): List of radiomics tables.

    Returns:
        pd.DataFrame: Single combined radiomics table.
    """
    # Initialization
    n_tables = len(rad_tables)

    base_idx = 0
    for idx, table in enumerate(rad_tables):
        if not table.empty:
            base_idx = idx
            break
    # Finding patient intersection
    for t in range(n_tables):
        if rad_tables[t].shape[1] > 0 and t != base_idx:
           rad_tables[base_idx], rad_tables[t] = intersect_var_tables(rad_tables[base_idx], rad_tables[t])
    
    # Check for NaNs
    '''for table in rad_tables:
        assert(table.isna().sum().sum() == 0)'''

    # Initializing the radiomics table template
    radiomics_table = pd.DataFrame()
    radiomics_table.Properties = {}
    radiomics_table._metadata += ['Properties']
    radiomics_table.Properties['userData'] = {}
    radiomics_table.Properties['VariableNames'] = []
    radiomics_table.Properties['userData']['normalization'] = {}

    # Combining radiomics table one by one
    count = 0
    continuous = []
    str_names = '||'
    for t in range(n_tables):
        rad_table_id = 'radTab' + str(t+1)
        if rad_tables[t].shape[1] > 0 and rad_tables[t].shape[0] > 0:
            features = rad_tables[t].columns.values
            description = rad_tables[t].Properties['Description']
            full_rad_names = get_full_rad_names(rad_tables[t].Properties['userData']['variables']['var_def'], 
                                                features)
            if 'normalization' in rad_tables[t].Properties['userData']:
                radiomics_table.Properties['userData']['normalization'][rad_table_id] = rad_tables[t].Properties[
                                                                                    'userData']['normalization']
            for f, feature in enumerate(features):
                count += 1
                var_name = 'radVar' + str(count)
                radiomics_table[var_name] = rad_tables[t][feature]
                radiomics_table.Properties['VariableNames'].append(var_name)
                continuous.append(var_name)
                if description:
                    str_names += 'radVar' + str(count) + ':' + description + '___' + full_rad_names[f] + '||'
                else:
                    str_names += 'radVar' + str(count) + ':' + full_rad_names[f] + '||'

    # Updating the radiomics table properties
    radiomics_table.Properties['Description'] = ''
    radiomics_table.Properties['DimensionNames'] = ['PatientID']
    radiomics_table.Properties['userData']['variables'] = {}
    radiomics_table.Properties['userData']['variables']['var_def'] = str_names
    radiomics_table.Properties['userData']['variables']['continuous'] = continuous

    return radiomics_table

def combine_tables_from_list(var_list: List, combination: List) -> pd.DataFrame:
    """
    Concatenates all variable tables in ``var_list`` according to ``var_ids``.
    
    Unlike ``combine_rad_tables`` This method concatenates variable tables instead of creating a new table from
    the intersection of the tables.

    Args:
        var_list (List): List of tables. Each key is a given var_id and holds a radiomic table.
            --> Ex: .var1: variable table 1
                    .var2: variable table 2
                    .var3: variable table 3
        combination (list): List of strings to identify the table to combine in var_list.
            --> Ex: {'var1','var3'}

    Returns:
        pd.DataFrame: variable_table: Combined radiomics table.
    """
    def concatenate_varid(var_names, var_id):
        return np.asarray([var_id + "__" + var_name for var_name in var_names.tolist()])
    
    # Initialization
    variables = dict()
    variables['continuous'] = np.array([])
    variable_tables = list()

    # Using the first table as template
    var_id = combination[0]
    variable_table = deepcopy(var_list[var_id]) # first table from the list
    variable_table.Properties = deepcopy(var_list[var_id].Properties)
    new_columns = [var_id + '__' + col for col in variable_table.columns]
    variable_table.columns = new_columns
    variable_table.Properties['VariableNames'] = new_columns
    variable_table.Properties['userData'] = dict()  # Re-Initializing
    variable_table.Properties['userData'][var_id] = deepcopy(var_list[var_id].Properties['userData'])
    variables['continuous'] = np.concatenate((variables['continuous'], var_list[var_id].Properties[
                                                                        'userData']['variables']['continuous']))
    variable_tables.append(variable_table)

    # Concatenating all other tables
    for var_id in combination[1:]:
        variable_table.Properties['userData'][var_id] = var_list[var_id].Properties['userData']
        patient_ids = intersect(list(variable_table.index), (var_list[var_id].index))
        var_list[var_id] = var_list[var_id].loc[patient_ids]
        variable_table = variable_table.loc[patient_ids]
        old_columns = list(variable_table.columns)
        old_properties = deepcopy(variable_table.Properties)  # for unknown reason Properties are erased after concat
        variable_table = pd.concat([variable_table, var_list[var_id]], axis=1)
        variable_table.columns = old_columns + [var_id + "__" + col for col in var_list[var_id].columns]
        variable_table.Properties = old_properties
        variable_table.Properties['VariableNames'] = list(variable_table.columns)
        variables['continuous'] = np.concatenate((variables['continuous'], var_list[var_id].Properties['userData']['variables']['continuous']))

    # Updating the radiomics table properties
    variable_table.Properties['Description'] = "Data table"
    variables['continuous'] = concatenate_varid(variables['continuous'], var_id)
    variable_table.Properties['userData']['variables'] = variables

    return variable_table

def convert_comibnations_to_list(combinations_string: str) -> Tuple[List, List]:
    """
    Converts a cell of strings specifying variable ids combinations to
    a cell of cells of strings.
    
    Args:
        combinations_string (str): Cell of strings specifying var_ids combinations
            separated by underscores.
            --> Ex: {'var1_var2';'var2_var3';'var1_var2_var3'}

    Rerturs:
        - List: List of strings of the seperated var_ids.
            --> Ex: {{'var1','var2'};{'var2','var3'};{'var1','var2','var3'}}
        - List: List of strings specifying the "alphabetical" IDs of combined variables 
            in ``combinations``. var1 --> A, var2 -> B, etc.
            --> Ex: {'model_AB';'model_BC';'model_ABC'}
    """
    # Building combinations
    combinations = [s.split('_') for s in combinations_string]

    # Building model_ids
    alphabet = string.ascii_uppercase
    model_ids = list()
    for combination in combinations:
        model_ids.append('model_' + ''.join([alphabet[int(var[3:])-1] for var in combination]))
    
    return combinations, model_ids

def count_class_imbalance(path_csv_outcomes: Path) -> Dict:
    """
    Counts the class imbalance in a given outcome table.
    
    Args:
        path_csv_outcomes (Path): Path to the outcome table.

    Returns:
        Dict: Dictionary containing the count of each class.
    """
    # Initialization
    outcomes = pandas.read_csv(path_csv_outcomes, sep=',')
    outcomes.dropna(inplace=True)
    outcomes.reset_index(inplace=True, drop=True)
    name_outcome = outcomes.columns[-1]
    
    # Counting the percentage of each class
    class_0_perc = np.sum(outcomes[name_outcome] == 0) / len(outcomes)
    class_1_perc = np.sum(outcomes[name_outcome] == 1) / len(outcomes)

    return {'class_0': class_0_perc, 'class_1': class_1_perc}

def create_experiment_folder(path_outcome_folder: str, method: str = 'Random') -> str:
    """
    Creates the experiment folder where the hold-out splits will be saved and returns the path
    to the folder.
    
    Args:
        path_outcome_folder (str): Full path to the outcome folder (folder containing the outcome table etc).
        method (str): String specifying the split type. Default is 'Random'.
    
    Returns:
        str: Full path to the experiment folder.
    """

    # Creating the outcome folder if it does not exist
    if not os.path.isdir(path_outcome_folder):
        os.makedirs(path_outcome_folder)

    # Creating the experiment folder if it does not exist
    list_outcome = os.listdir(path_outcome_folder)
    if not list_outcome:
        flag_exist_split = False
    else:
        n_exist = 0
        flag_exist_split = False
        for i in range(len(list_outcome)):
            if 'holdOut__' + method + '__' in list_outcome[i]:
                n_exist = n_exist + 1
                flag_exist_split = True
    
    # If path experiment folder exists already, create a new one (sequentially)
    if not flag_exist_split:
        path_split = str(path_outcome_folder) + '/holdOut__' + method + '__001'
    else:
        path_split = str(path_outcome_folder) + '/holdOut__' + method + '__' + \
            str(n_exist+1).zfill(3)

    os.mkdir(path_split)
    return path_split

def create_holdout_set(
                path_outcome_file: Union[str, Path],
                outcome_name: str,
                path_save_experiments: Union[str, Path] = None,
                method: str = 'random',
                percentage: float = 0.2,
                n_split: int = 1,
                seed : int = 1) -> None:
    """
    Creates a hold-out patient set to be used for final independent testing after a final 
    model is chosen. All the information is saved in a JSON file.
    
    Args:
        path_outcome_file (str): Full path to where the outcome CSV file is stored.
        outcome_name (str): Name of the outcome. For example, 'OS' for overral survivor.
        path_save_experiments (str): Full path to the folder where the experiments
            will be saved.
        method (str): Method to use for creating the hold-out set. Options are:
            - 'random': Randomly selects patients for the hold-out set.
            - 'all_learn': No hold-out set is created. All patients are used for learning.
            - 'institution': TODO.
        percentage (float): Percentage of patients to use for the hold-out set. Default is 0.2.
        n_split (int): Number of splits to create. Default is 1.
        seed (int): Seed to use for the random split. Default is 1.
    
    Returns:
        None.
    """
    # Initilization
    outcome_name = outcome_name.upper()
    outcome_table = pandas.read_csv(path_outcome_file, sep=',')
    outcome_table.dropna(inplace=True)
    outcome_table.reset_index(inplace=True, drop=True)
    patient_ids = outcome_table['PatientID']

    # Creating experiment folders and patient test split(s)
    outcome_name = re.sub(r'\W', "", outcome_name)
    path_outcome = str(path_save_experiments) + '/' + outcome_name
    name_outcome_in_table_binary = outcome_name + '_binary'
    
    # Column names in the outcome table
    with open(path_outcome_file, 'r') as infile:
        reader = csv.DictReader(infile, delimiter=',')
        var_names = reader.fieldnames

    # Include time to event if it exists
    flag_time = False
    if(outcome_name + '_eventFreeTime' in str(var_names)):
        name_outcome_in_table_time = outcome_name + '_eventFreeTime'
        flag_time = True

    # Check if the outcome name for binary is correct
    if name_outcome_in_table_binary not in outcome_table.columns:
            name_outcome_in_table_binary = var_names[-1]
    
    # Run the split
    # Random
    if 'random' in method.lower():
        # Creating the experiment folder
        path_split = create_experiment_folder(path_outcome, 'random')

        # Getting the random split
        patients_learn_temp, patients_hold_out_temp = get_stratified_splits(
            outcome_table[['PatientID', name_outcome_in_table_binary]],
            n_split, percentage, seed, False)
        
        # Getting the patient IDs in the learning and hold-out sets
        if n_split > 1:
            patients_learn = np.empty((n_split, len(patients_learn_temp[0])), dtype=object)
            patients_hold_out = np.empty((n_split, len(patients_hold_out_temp[0])), dtype=object)
            for s in range(n_split):
                patients_learn[s] = patient_ids[patients_learn_temp[s]]
                patients_hold_out[s] = patient_ids[patients_hold_out_temp[s]]
        else:
            patients_learn = patient_ids[patients_learn_temp.values.tolist()]
            patients_learn.reset_index(inplace=True, drop=True)
            patients_hold_out = patient_ids[patients_hold_out_temp.values.tolist()]
            patients_hold_out.reset_index(inplace=True, drop=True)
    
    # All Learn
    elif 'all_learn' in method.lower():
        # Creating the experiment folder
        path_split = create_experiment_folder(path_outcome, 'all_learn')

        # Getting the split (all Learn so no hold out)
        patients_learn = patient_ids
        patients_hold_out = []
    else:
        raise ValueError('Method not recognized. Use "random" or "all_learn".')
    
    # Creating final outcome table and saving it
    if flag_time:
        outcomes = outcome_table[
            ['PatientID', name_outcome_in_table_binary, name_outcome_in_table_time]]
    else:
        outcomes = outcome_table[['PatientID', name_outcome_in_table_binary]]

    # Finalize the outcome table
    outcomes = outcomes.dropna(inplace=False)   # Drop NaNs
    outcomes.reset_index(inplace=True, drop=True)   # Reset index        

    # Save the outcome table
    paths_exp_outcomes = str(path_split + '/outcomes.csv')
    outcomes.to_csv(paths_exp_outcomes, index=False)

    # Save dict of patientsLearn
    paths_exp_patientsLearn = str(path_split) + '/patientsLearn.json'
    patients_learn.to_json(paths_exp_patientsLearn, orient='values', indent=4)

    # Save dict of patientsHoldOut
    if method == 'random':
        paths_exp_patients_hold_out = str(path_split) + '/patientsHoldOut.json'
        patients_hold_out.to_json(paths_exp_patients_hold_out, orient='values', indent=4)

        # Save dict of all the paths
        data={
            "outcomes" : paths_exp_outcomes,
            "patientsLearn": paths_exp_patientsLearn,
            "patientsHoldOut": paths_exp_patients_hold_out,
            "pathWORK": path_split
        }
    else:
        data={
            "outcomes" : paths_exp_outcomes,
            "patientsLearn": paths_exp_patientsLearn,
            "pathWORK": path_split
        }
    paths_exp = str(path_split + '/paths_exp.json')
    with open(paths_exp, 'w') as f:
        json.dump(data, f, indent=4)
    
    # Return the path to the experiment and path to split
    return path_split, paths_exp

def cross_validation_split(
        outcome: List[Union[int, float]], 
        n_splits: int = 5, 
        seed: int = None
    ) -> Tuple[List[List[int]], List[List[int]]]:
    """
    Perform stratified cross-validation split.

    Args:
        outcome (list): Outcome variable (binary).
        n_splits (int, optional): Number of folds. Default is 5.
        seed (int or None, optional): Random seed for reproducibility. Default is None.

    Returns:
        train_indices_list (list of lists): List of training indices for each fold.
        test_indices_list (list of lists): List of testing indices for each fold.
    """

    skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=seed)
    train_data_list = []
    test_data_list = []
    patient_ids = pd.Series(outcome.index)

    for train_indices, test_indices in skf.split(X=outcome, y=outcome):
        train_data_list.append(patient_ids[train_indices])
        test_data_list.append(patient_ids[test_indices])
    
    train_data_array = np.array(train_data_list, dtype=object)
    test_data_array = np.array(test_data_list, dtype=object)

    return train_data_array, test_data_array

def find_best_model(path_results: Path, metric: str = 'AUC', second_metric: str = 'AUC') -> Tuple[Dict, Path]:
    """
    Find the best model with the highest performance on the test set
    in a given path based on a given metric.

    Args:
        path_results (Path): Path to the results folder.
        metric (str): Metric to use to find the best model in case of a tie. Default is 'AUC'.

    Returns:
        Tuple[Dict, Path]: Tuple containing the best model result dict and the path to the best model.
    """
    list_metrics = [
        'AUC', 'Sensitivity', 'Specificity', 
        'BAC', 'AUPRC', 'Precision', 
        'NPV', 'Accuracy', 'F1_score', 'MCC',
        'TP', 'TN', 'FP', 'FN'
    ]
    assert metric in list_metrics, f'Given metric {metric} is not in the list of metrics. Please choose from {list_metrics}'

    # Get all tests paths
    list_path_tests =  [path for path in path_results.iterdir() if path.is_dir()]

    # Initialization
    metric_best = -1
    second_metric_best = -1
    path_result_best = None

    # Get all models and their metrics (AUC especially)
    for path_test in list_path_tests:
        if not (path_test / 'run_results.json').exists():
            continue
        results_dict = load_json(path_test / 'run_results.json')
        metric_test = results_dict[list(results_dict.keys())[0]]['test']['metrics'][metric]
        if metric_test > metric_best:
            metric_best = metric_test
            path_result_best = path_test
        elif metric_test == metric_best:
            second_metric_test = results_dict[list(results_dict.keys())[0]]['test']['metrics'][second_metric]
            if second_metric_test > second_metric_best:
                second_metric_best = second_metric_test
                path_result_best = path_test
    
    # Load best model result dict
    results_dict_best = load_json(path_result_best / 'run_results.json')

    # Load model
    model_name = list(results_dict_best.keys())[0]
    with open(path_result_best / f'{model_name}.pickle', 'rb') as file:
        model = pickle.load(file)
    
    return model, results_dict_best

def feature_imporance_analysis(path_results: Path):
    """
    Averages the results (AUC, BAC, Sensitivity and Specifity) of all the runs of the same experiment,
    for training, testing and holdout sets.

    Args:
        path_results(Path): path to the folder containing the results of the experiment.
        save (bool, optional): If True, saves the results in the same folder as the model.
    
    Returns:
        None.
    """
    # Get all tests paths
    list_path_tests =  [path for path in path_results.iterdir() if path.is_dir()]

    # Initialization
    results_avg_temp = {}
    results_avg = {}

    # Process metrics
    for path_test in list_path_tests:
        variables = []
        list_models = list(path_test.glob('*.pickle'))
        if len(list_models) == 0 or len(list_models) > 1:
            raise ValueError(f'Path {path_test} does not contain a single model.')
        model_obj = list_models[0]
        with open(model_obj, "rb") as f:
            model_dict = pickle.load(f)
        if model_dict["var_names"]:
            variables = get_full_rad_names(model_dict['var_info']['variables']['var_def'], model_dict["var_names"])
        for index, var in enumerate(variables):
            var = var.split("\\")[-1]   # Remove the path for windows
            var = var.split("/")[-1]    # Remove the path for linux
            if var not in results_avg_temp:
                results_avg_temp[var] = {
                    'importance_mean': [],
                    'times_selected': 0
                }
            
            results_avg_temp[var]['importance_mean'].append(model_dict['model'].feature_importances_[index])
            results_avg_temp[var]['times_selected'] += 1
    for var in results_avg_temp:
        results_avg[var] = {
            'importance_mean': np.sum(results_avg_temp[var]['importance_mean']) / len(list_path_tests),
            'times_selected': results_avg_temp[var]['times_selected']
        }
    
    del results_avg_temp
            
    save_json(path_results / 'feature_importance_analysis.json', results_avg, cls=NumpyEncoder)

def get_ml_test_table(variable_table: pd.DataFrame, var_names: List, var_def: str) -> pd.DataFrame:
    """
    Gets the test table with the variables that are present in the training table.

    Args:
        variable_table (pd.DataFrame): Table with the variables to use for the ML model that 
            will be matched with the training table.
        var_names (List): List of variable names used for the ML model .
        var_def (str): String of the full variables names used for the ML model.
    
    Returns:
        pd.DataFrame: Table with the variables that are present in the training table.
    """

    # Get the full variable names for training
    full_radvar_names_trained = get_full_rad_names(var_def, var_names).tolist()

    # Get the full variable names for testing
    full_rad_var_names_test = get_full_rad_names(
        variable_table.Properties['userData']['variables']['var_def'], 
        variable_table.columns.values
    ).tolist()

    # Get the indexes of the variables that are present in the training table
    indexes = []
    for radvar in full_radvar_names_trained:
        try:
            indexes.append(full_rad_var_names_test.index(radvar))
        except ValueError as e:
            print(e)
            raise ValueError('The variable ' + radvar + ' is not present in the test table.')

    # Get the test table with the variables that are present in the training table
    variable_table = variable_table.iloc[:, indexes]

    # User data - var_def
    str_names = '||'
    for v in range(len(var_names)):
        str_names += var_names[v] + ':' + full_radvar_names_trained[v] + '||'

    # Update metadata and variable names
    variable_table.columns = var_names
    variable_table.Properties['VariableNames'] = var_names
    variable_table.Properties['userData']['variables']['var_def'] = str_names
    variable_table.Properties['userData']['variables']['continuous'] = var_names
    
    # Rename columns to s sequential names again
    return variable_table

def finalize_rad_table(rad_table: pd.DataFrame) -> pd.DataFrame:
    """
    Finalizes the variable names and the associated metadata. Used to have sequential variable 
    names and UserData with only variable names present in the table.
    
    Args:
        rad_table (pd.DataFrame): radiomics table to be finalized.
    
    Returns:
        pd.DataFrame: Finalized radiomics table.
    """

    # Initialization
    var_names = rad_table.columns.values
    full_rad_names = get_full_rad_names(rad_table.Properties['userData']['variables']['var_def'], var_names)

    # User data - var_def
    str_names = '||'
    for v in range(var_names.size):
        var_names[v] = 'radVar' + str(v+1)
        str_names = str_names + var_names[v] + ':' + full_rad_names[v] + '||'

    # Update metadata and variable names
    rad_table.columns = var_names
    rad_table.Properties['VariableNames'] = var_names
    rad_table.Properties['userData']['variables']['var_def'] = str_names
    rad_table.Properties['userData']['variables']['continuous'] = var_names

    return rad_table

def get_radiomics_table(
        path_radiomics_csv: Path, 
        path_radiomics_txt: Path, 
        image_type: str, 
        patients_ids: List = None
    ) -> pd.DataFrame:
    """
    Loads the radiomics table from the .csv file and the associated metadata.
    
    Args:
        path_radiomics_csv (Path): full path to the csv file of radiomics table.
            --> Ex: /home/myStudy/FEATURES/radiomics__PET(GTV)__image.csv
        path_radiomics_txt: full path to the radiomics variable definitions in text format (associated
            to path_radiomics_csv).
            -> Ex: /home/myStudy/FEATURES/radiomics__PET(GTV)__image.txt
        image_type (str): String specifying the type of image on which the radiomics
            features were computed.
            --> Format: $scan$($roiType$)__$imSpace$
            --> Ex: PET(tumor)__HHH_coif1
        patients_ids (list, optional): List of strings specifying the patientIDs of
            patients to fetch from the radiomics table. If this
            argument is not present, all patients are fetched.
            --> Ex: {'Cervix-UCSF-001';Cervix-McGill-004}

    Returns:
        pd.DataFrame: radiomics table
    """
    # Read CSV table
    radiomics_table = pd.read_csv(path_radiomics_csv, index_col=0)
    if patients_ids is not None:
        patients_ids = intersect(patients_ids, list(radiomics_table.index))
        radiomics_table = radiomics_table.loc[patients_ids]

    # Read the associated TXT file
    with open(path_radiomics_txt, 'r') as f:
        user_data = f.read()

    # Grouping the information
    radiomics_table._metadata += ["Properties"]
    radiomics_table.Properties = dict()
    radiomics_table.Properties['userData'] = dict()
    radiomics_table.Properties['userData']['variables'] = dict()
    radiomics_table.Properties['userData']['variables']['var_def'] = user_data
    radiomics_table.Properties['Description'] = image_type

    # Only continuous will be used for now but this design will facilitate the use of 
    # other categories in the future.
    # radiomics = continous.
    radiomics_table.Properties['userData']['variables']['continuous'] = np.asarray(list(radiomics_table.columns.values))

    return radiomics_table

def get_splits(outcome: pd.DataFrame, n_split: int, test_split_proportion: float) -> Tuple[List, List]:
    """
    Splits the given outcome table in two sets.

    Args:
        outcome (pd.DataFrame): Table with a single outcome column of 0's and 1's.
        n_splits (int): Integer specifying the number of splits to create.
        test_split_proportion (float): Float between 0 and 1 specifying the proportion
                of patients to include in the test set.

    Returns:
        train_sets List of indexes for the train_sets.
        test_sets: List of indexes for the test_sets.

    """

    ind_neg = np.where(outcome == 0)
    n_neg = len(ind_neg[0])
    ind_pos = np.where(outcome == 1)
    n_pos = len(ind_pos[0])
    n_neg_test = round(test_split_proportion * n_neg)
    n_pos_test = round(test_split_proportion * n_pos)

    n_inst = len(outcome)
    n_test = n_pos_test + n_neg_test
    n_train = n_inst - n_test

    if(n_split==1):
        train_sets = np.zeros(n_train)
        test_sets = np.zeros(n_test)
    else:
        train_sets = np.zeros((n_split, n_train))
        test_sets = np.zeros((n_split, n_test))

    for s in range(n_split):
        ind_pos_test = np.random.choice(ind_pos[0], n_pos_test, replace=False)
        ind_neg_test = np.random.choice(ind_neg[0], n_neg_test, replace=False)

        ind_test = np.concatenate((ind_pos_test,ind_neg_test))
        ind_test.sort()

        ind_train = np.arange(n_inst)
        ind_train = np.delete(ind_train, ind_test)
        ind_train.sort()

        if(n_split>1):
            train_sets[s] = ind_train
            test_sets[s] = ind_test
        else:
            train_sets = ind_train
            test_sets = ind_test

    return train_sets, test_sets

def get_stratified_splits(
        outcome_table: pd.DataFrame,
        n_splits: int,
        test_split_proportion: float,
        seed: int,
        flag_by_cat: bool=False
    ) -> Tuple[List, List]:
    """
    Sub-divides a given outcome dataset into multiple stratified patient splits. 
    The stratification is performed per class proportion (or by institution).

    Args:
        outcome_table: Table with a single outcome column of 0's and 1's.
                    The rows of the table must define the patient IDs: $Cancer-$Institution-$Number.
        n_splits: Integer specifying the number of splits to create.
        test_split_proportion: Float between 0 and 1 specifying the proportion 
            of patients to include in the test set.
        seed: Integer specifying the random generator seed to use for random splitting.
        flag_by_cat (optional): Logical flag specifying if we are to produce 
            the split by taking into account the institutions in the outcome table. 
            If true, patients in Training and testing splits have the same prortion 
            of events per instiution as originally found in the initial data. Default: False.

    Returns:
        List: patients_train_splits, list of size nTrainXnSplit, where each entry
            is a string specifying a "Training" patient.
        List: patients_test_splits, list of size nTestXnSplit, where each entry
            is a string specifying a "testing" patient
    """
    patient_ids = pd.Series(outcome_table.index)
    patients_train_splits = []
    patients_test_splits = []

    # Take into account the institutions in the outcome table
    if flag_by_cat:
        institution_cat_vector = get_institutions_from_ids(patient_ids)
        all_categories = np.unique(institution_cat_vector)
        n_cat = len(all_categories)
        # Split for each institution
        for i in range(n_cat):
            np.random.seed(seed)
            cat = all_categories[i]
            flag_cat = institution_cat_vector == cat
            patient_ids_cat = patient_ids[flag_cat]
            patient_ids_cat.reset_index(inplace=True, drop=True)

            # Split train and test sets
            train_sets, test_sets = get_splits(outcome_table[flag_cat.values], n_splits, test_split_proportion)

            if n_splits > 1:
                temp_patients_train = np.empty((n_splits, len(train_sets[0])), dtype=object)
                temp_patientsTest = np.empty((n_splits, len(test_sets[0])), dtype=object)
                for s in range(n_splits):
                    temp_patients_train[s] = patient_ids_cat[train_sets[s]]
                    temp_patientsTest[s] = patient_ids_cat[test_sets[s]]
            else:
                temp_patients_train = patient_ids_cat[train_sets]
                temp_patients_train.reset_index(inplace=True, drop=True)
                temp_patientsTest = patient_ids_cat[test_sets]
                temp_patientsTest.reset_index(inplace=True, drop=True)
            
            # Initialize the train and test patients list (1st iteration)
            if i==0:
                patients_train_splits=temp_patients_train
                patients_test_splits=temp_patientsTest

            # Add new patients to the train and test patients list (other iterations)
            if i>0:
                if n_splits>1:
                    patients_train_splits = np.append(patients_train_splits, temp_patients_train, axis=1)
                    patients_test_splits = np.append(patients_test_splits, temp_patientsTest, axis=1)

                else: 
                    patients_train_splits = np.append(patients_train_splits, temp_patients_train)
                    patients_test_splits = np.append(patients_test_splits, temp_patientsTest)
    
    # Do not take into account the institutions in the outcome table
    else:
        # Split train and test sets
        train_sets, test_sets = get_splits(outcome_table, n_splits, test_split_proportion)
        if n_splits > 1:
            patients_train_splits = np.empty((n_splits, len(train_sets[0])), dtype=object)
            patients_test_splits = np.empty((n_splits, len(test_sets[0])), dtype=object)
            for s in range(n_splits):
                patients_train_splits[s] = patient_ids[train_sets[s]]
                patients_test_splits[s] = patient_ids[test_sets[s]]
        else:
            patients_train_splits = patient_ids[train_sets]
            patients_train_splits.reset_index(inplace=True, drop=True)
            patients_test_splits = patient_ids[test_sets]
            patients_test_splits.reset_index(inplace=True, drop=True)

    return patients_train_splits, patients_test_splits

def get_patient_id_classes(outcome_table: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Yields the patients from the majority class and the minority class in the given outcome table.
    Only supports binary classes.
    
    Args:
        outcome_table(pd.DataFrame): outcome table with binary labels.
    
    Returns:
        pd.DataFrame: Majority class patientIDs.
        pd.DataFrame: Minority class patientIDs.
    """
    ones = outcome_table.loc[outcome_table.iloc[0:].values == 1].index
    zeros = outcome_table.loc[outcome_table.iloc[0:].values == 0].index
    if ones.size > zeros.size:
        return ones, zeros
    
    return zeros, ones

def intersect(list1: List, list2: List, sort: bool = False) -> List:
    """
    Returns the intersection of two list.

    Args:
        list1 (List): the first list.
        list2 (List): the second list.
        order (bool): if True, the intersection is sorted.
    
    Returns:
        List: the intersection of the two lists.
    """

    intersection = list(filter(lambda x: x in list1, list2))
    if sort:
        return sorted(intersection)
    return intersection

def intersect_var_tables(var_table1: pd.DataFrame, var_table2: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    This function takes 2 variable table, compares the indexes and drops the
    ones that are not in both, then returns the 2 table.

    Args:
        var_table1 (pd.DataFrame): first variable table.
        var_table2 (pd.DataFrame): second variable table.
    
    Returns:
        pd.DataFrame: first variable table with the same indexes as the second.
        pd.DataFrame: second variable table with the same indexes as the first.
    """
    # Find the unique values in var_table1 that are not in var_table2
    missing = np.setdiff1d(var_table1.index.to_numpy(), var_table2.index.to_numpy())
    if missing.size > 0:
        var_table1 = var_table1.drop(missing)

    # Find the unique values in var_table2 that are not in var_table1
    missing = np.setdiff1d(var_table2.index.to_numpy(), var_table1.index.to_numpy())
    if missing.size > 0:
        var_table2 = var_table2.drop(missing)

    return var_table1, var_table2

def under_sample(outcome_table_binary: pd.DataFrame) -> pd.DataFrame:
    """
    Performs under-sampling to obtain an equal number of outcomes in the binary outcome table.
    
    Args:
        outcome_table_binary (pd.DataFrame): outcome table with binary labels.
    
    Returns:
        pd.DataFrame: outcome table with balanced binary labels.
    """

    # We place them prematurely in maj and min and correct it afterwards
    n_maj = (outcome_table_binary == 0).sum().values[0]
    n_min = (outcome_table_binary == 1).sum().values[0]
    if n_maj == n_min:
        return outcome_table_binary
    elif n_min > n_maj:
        n_min, n_maj = n_maj, n_min

    # Sample the patients from the majority class
    patient_ids_maj, patient_ids_min = get_patient_id_classes(outcome_table_binary)
    patient_ids_min = list(patient_ids_min)
    patient_ids_numpy = patient_ids_maj.to_numpy()
    np.random.shuffle(patient_ids_numpy)
    patient_ids_sample = list(patient_ids_numpy[0:n_min])
    new_ids = patient_ids_min + patient_ids_sample

    return outcome_table_binary.loc[new_ids, :]

def save_model(model: Dict, var_id: str, path_model: Path, ml: Dict = None, name_type: str = "") -> Dict:
    """
    Saves a given model locally as a pickle object and outputs a dictionary
    containing the model's information.

    Args:
        model (Dict): The model dict to save.
        var_id (str): The stduied variable. For ex: 'var3'.
        path_model (str): The path to save the model.
        ml (Dict, optional): Dicionary containing the settings of the machine learning experiment.
        name_type (str, optional): String specifying the type of the variable. For examlpe: "RadiomicsIntensity". Default is "".
    
    Returns:
        Dict: A dictionary containing the model's information.
    """
    # Saving model
    with open(path_model, "wb") as f:
        pickle.dump(model, f)

    # Getting the "var_names" string
    if ml is not None:
        var_names = ml['variables'][var_id]['nameType']
    elif name_type != "":
        var_names = name_type
    else:
        var_names = [var_id]

    # Recording model info
    model_info = dict()
    model_info['path'] = path_model
    model_info['var_ids'] = var_id
    model_info['var_type'] = var_names

    try: # This part may fail if model training failed.
        model_info['var_names'] = model['var_names']
        model_info['var_info'] = model['var_info']
        if 'normalization' in model_info['var_info'].keys():
            if 'normalization_table' in model_info['var_info']['normalization'].keys():
                normalization_struct = write_table_structure(model_info['var_info']['normalization']['normalization_table'])
                model_info['var_info']['normalization']['normalization_table'] = normalization_struct
        model_info['threshold'] = model['threshold']
    except Exception as e:
        print("Failed to create a fully model info")
        print(e)

    return model_info

def write_table_structure(data_table: pd.DataFrame) -> Dict:
    """
    Writes the structure of a table in a dictionary.

    Args:
        data_table (pd.DataFrame): a table.
    
    Returns:
        Dict: a dictionary containing the table's structure.
    """
    # Initialization
    data_struct = dict()

    if len(data_table.index) != 0:
        data_struct['index'] = list(data_table.index)

    # Creating the structure
    for column in data_table.columns:
        data_struct[column] = data_table[column]

    return data_struct



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import json
import pathlib
from typing import Dict


def _is_jsonable(data: any, cls: object) -> bool:
    """Checks if the given ``data`` is JSON serializable.

    Args:
        data (Any): ``Data`` that will be checked.
        cls(object, optional): Costum JSONDecoder subclass. If not specified JSONDecoder is used. 

    Returns:
        bool: True if the given ``data`` is serializable, False if not.
    """
    try:
        json.dumps(data, cls=cls)
        return True
    except (TypeError, OverflowError):
        return False


def posix_to_string(dictionnary: Dict) -> Dict:
    """Converts all Pathlib.Path to str [Pathlib is not serializable].

    Args:
        dictionnary (Dict):  dict with Pathlib.Path values to convert.

    Returns:
        Dict: ``dictionnary`` with all Pathlib.Path converted to str.
    """
    for key, value in dictionnary.items():
        if type(value) is dict:
            value = posix_to_string(value)
        else:
            if issubclass(type(value), (pathlib.WindowsPath, pathlib.PosixPath, pathlib.Path)):
               dictionnary[key] = str(value)
    
    return dictionnary

def load_json(file_path: pathlib.Path) -> Dict:
    """Wrapper to json.load function.

    Args:
        file_path (Path): Path of the json file to load.

    Returns:
        Dict: The loaded json file.
     
    """
    with open(file_path, 'r') as fp:
        return json.load(fp)


def save_json(file_path: pathlib.Path, data: any, cls=None) -> None:
    """Wrapper to json.dump function.

    Args:
        file_path (Path): Path to write the json file to.
        data (Any): Data to write to the given path. Must be serializable by JSON.
        cls(object, optional): Costum JSONDecoder subclass. If not specified JSONDecoder is used. 

    Returns:
        None: saves the ``data`` in JSON file to the ``file_path``.

    Raises:
        TypeError: If ``data`` is not JSON serializable.
    """
    if _is_jsonable(data, cls):
        with open(file_path, 'w') as fp:
            json.dump(data, fp, indent=4, cls=cls)
    else:
        raise TypeError("The given data is not JSON serializable. \
            We rocommend using a costum encoder.")
